using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
namespace Theremino_HAL
{



	// ==================================================================================================
	//   CLASS MASTER
	// ==================================================================================================

	internal partial class Master
	{

		private object ActionLock = new object();
		private string Name;

		private Int32 mSlaveCount;
        internal Theremino_HID Hid;
		internal Int32 ConfigId;
		internal bool ConfigValid;
		internal List<Slave> Slaves;
		internal Int32 MasterId;
		
		internal Int32 CommSpeed;
		internal bool FastComm;
		internal float CommMillisec;
		internal float CommFps;

		internal float ErrorRate;

		internal Int32 SlaveCount {
			get { return mSlaveCount; }
		}

		internal void SetName(string _name)
		{
			Name = _name;
		}

		internal string GetName()
		{
			return Name;
		}

		internal Master(Int32 _masterid)
		{
			Name = "NoName";
			MasterId = _masterid;
			mSlaveCount = 0;
			CommSpeed = 7;
			FastComm = false;
			CommFps = 100;
			Slaves = new List<Slave>();
		}

		//Protected Overrides Sub Finalize()
        //    If Not Module_SaveLoad.CanUseWindowsDlls() Then
		//        If Hid.HidDevice_Unix.IsHidOpen() Then
		//            Hid.HidDevice_Unix.HidClose()
		//        End If
		//    End If
		//End Sub



		// ==========================================================================================
		//   HOST TO MASTER COMMANDS
		// ==========================================================================================

		internal enum Cmd : byte
		{
			// --------------------------------------- commands that master expands for the slaves
			RECOG_START = 254,
			//CMD_RECOG 						= 253 ( Master to Slaves only )
			FAST_DATA_EXCHANGE = 251,
			//
			// --------------------------------------- commands that master repeats to slaves
			SETUP_SLAVE_PINS = 249,
			SET_MASTER_NAME = 248,
			GET_MASTER_NAME = 247,
			SEND_VALUES = 246,
			GET_VALUES = 245,
			SEND_BYTES = 244,
			GET_BYTES = 243,
			//
			// --------------------------------------- commands to master only  ( < 200 )
			SPEED = 199,
			NOCMD = 0
		}




		// ==========================================================================================
		//   PUBLIC COMMANDS
		// ==========================================================================================

		internal void Calibrate()
		{
			if (!CalibrationNeeded()) return;
 
			DataExchangeThread_Stop();
			lock (ActionLock) {
                Module_Utils.SleepMyThread(10);
				DataExchange();
                Module_Utils.SleepMyThread(10);
				for (Int32 i = 0; i <= SlaveCount - 1; i++) {
                    Slaves[i].Calibrate();
				}
			}
			DataExchangeThread_Start();
		}

		internal void CalibrateRawZero()
		{
			for (Int32 i = 0; i <= SlaveCount - 1; i++) {
				Slaves[i].CalibrateRawZero();
			}
		}

        internal bool CalibrationNeeded()
        {
            for (Int32 i = 0; i <= SlaveCount - 1; i++)
            {
                if (Slaves[i].CalibrationNeeded())
                {
                    return true;
                }
            }
            return false;
        }

		internal void SetSpeed()
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				SendSpeedToSlaves();
			}
			DataExchangeThread_Start();
		}

		internal void SetSpeed(Int32 _Speed)
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				CommSpeed = _Speed;
				SendSpeedToSlaves();
			}
			DataExchangeThread_Start();
		}



		internal void Recognize()
		{
			//' ------------------------------------------------- try to unlock slaves
			//Dim b(60) As Byte
			//SendBytesToSlave(255, b)
            //Module_Utils.SleepMyThread(100)
			//' -------------------------------------------------

			DataExchangeThread_Stop();
			lock (ActionLock) {
				Slaves.Clear();
				// --------------------------------------------- SPEED - TODO provare con molti slave e lasciarne solo due
				SendSpeedToSlaves();
				SendSpeedToSlaves();
				//SendSpeedToSlaves()
				//SendSpeedToSlaves()
				//SendSpeedToSlaves()
				//SendSpeedToSlaves()
				// --------------------------------------------- DETECT
				Hid.USB_TxData[0] = (byte)Cmd.RECOG_START;
				Hid.USB_TxData[1] = 0;
				// --------------------------------------------- WRITE-READ
				Hid.USB_Write_Read();
                Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
				// --------------------------------------------- ERROR ?
				if (Hid.USB_RxData[0] != 0) {
					Execution_Error("Recognize");
				}
				else {
					Execution_OK();
				}
				// --------------------------------------------- INIT 
				mSlaveCount = Hid.USB_RxData[1];
				for (Int32 i = 0; i <= SlaveCount - 1; i++) {
					Slave.SlaveTypes s = (Slave.SlaveTypes)Hid.USB_RxData[i + 2];
					switch (s) {
						case Slave.SlaveTypes.CapSensor:
							Slaves.Add(new Slave_CapSensor(MasterId, i));
							break;
						case Slave.SlaveTypes.Servo:
							Slaves.Add(new Slave_Servo(MasterId, i));
							break;
						case Slave.SlaveTypes.Generic:
							Slaves.Add(new Slave_Generic(MasterId, i));
							break;
						case Slave.SlaveTypes.InOut:
							Slaves.Add(new Slave_InOut(MasterId, i));
							break;
						case Slave.SlaveTypes.MasterPins:
							Slaves.Add(new Slave_MasterPins(MasterId, i));
							break;
						case Slave.SlaveTypes.RmsLogger:
							Slaves.Add(new Slave_RmsLogger(MasterId, i));
							break;
						case Slave.SlaveTypes.Sniffer:
							Slaves.Add(new Slave_Sniffer(MasterId, i));
							break;
						default:
							mSlaveCount = 0;
							Slaves.Clear();
							break;
					}
				}
			}
			DataExchangeThread_Start();
		}

		internal void SetupSlavePins(Int32 slaveID)
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				Slave s = Slaves[slaveID];
                Hid.USB_TxData[0] = (byte)Cmd.SETUP_SLAVE_PINS;
				Hid.USB_TxData[1] = (byte)slaveID;
				Hid.USB_TxData[2] = (byte)s.Pins.Count;
				for (Int32 i = 0; i <= s.Pins.Count - 1; i++) 
                {
					Hid.USB_TxData[i + 3] = (byte)s.Pins[i].GetPinType();
				}
				Hid.USB_Write_Read();
                Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
				// --------------------------------------------- error ?
				if (Hid.USB_RxData[0] != 0) {
					Execution_Error("SetupSlavePins");
				}
				else {
					Execution_OK();
				}
				// ---------------------------------------------
				SetCommunicationIndexesToPins();
			}
			DataExchangeThread_Start();
		}

		internal void WriteNameToHardware()
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				Name = Strings.Left(Name, 16);
                Hid.USB_TxData[0] = (byte)Cmd.SET_MASTER_NAME;
				Int32 i = default(Int32);
				for (i = 1; i <= Name.Length; i++) {
					Hid.USB_TxData[i] = (byte)Strings.Asc(Strings.Mid(Name, i, 1));
				}
				Hid.USB_TxData[i] = (byte)0;
				Hid.USB_Write_Read();
                Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
				// --------------------------------------------- error ?
				if (Hid.USB_RxData[0] != 0) {
					Execution_Error("WriteNameToHardware");
				}
				else {
					Execution_OK();
				}
			}
			DataExchangeThread_Start();
		}

		internal void ReadNameFromHardware()
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				Name = "";
                Hid.USB_TxData[0] = (byte)Cmd.GET_MASTER_NAME;
				Hid.USB_Write_Read();
                Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
				// --------------------------------------------- error ?
				if (Hid.USB_RxData[0] == 0) {
					Int32 i = 1;
					do {
						if (Hid.USB_RxData[i] == 0) break; // TODO: might not be correct. Was : Exit Do
 
						Name += Strings.Chr(Hid.USB_RxData[i]);
						i += 1;
					}
					while (true);
					Execution_OK();
				}
				else {
					Execution_Error("ReadNameToHardware");
				}
			}
			DataExchangeThread_Start();
		}



		internal void SetOutputPinsToParkingPosition()
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				foreach (Slave s in Slaves) {
					foreach (Pin p in s.Pins) {
						if (p.Direction == Pin.Directions.HostToMaster) {
							p.Value = p.Value_Parking;
							p.ConvertValueToHardwareFormat();
						}
					}
				}
				// ------------------------------------------------- USB <> Master <> Slaves
				DataExchange();
                Module_Utils.SleepMyThread(32);
			}
			DataExchangeThread_Start();
		}

		internal void StopAllSnifferLogs()
		{
			foreach (Slave s in Slaves) {
				foreach (Pin p in s.Pins) {
					p.SnifferLogEnabled = false;
				}
			}
		}


		internal void SetCommunicationIndexesToPins()
		{
			// ---------------------------------------------------- byte 0 is the command
			Int32 UsbByteIndex_HostToHardware = 1;
			// ---------------------------------------------------- byte 0 is the response
			Int32 UsbByteIndex_HardwareToHost = 1;
			//
			foreach (Slave s in Slaves) {
				foreach (Pin p in s.Pins) {
					// --------------------------------------------------- USB communication byte index
					switch (p.Direction) {
						case Pin.Directions.HostToMaster:
							p.UsbByteIndex = UsbByteIndex_HostToHardware;
							UsbByteIndex_HostToHardware += p.UsbBytesCount;
							break;
						case Pin.Directions.MasterToHost:
							p.UsbByteIndex = UsbByteIndex_HardwareToHost;
							UsbByteIndex_HardwareToHost += p.UsbBytesCount;
							break;
					}
					//Debug.Print(p.PinId.ToString & "  " & Pin.PinTypeToString(p.GetPinType) & "  " & p.UsbByteIndex)
				}
			}
		}



		// ==========================================================================================
		//   INTERNAL COMMANDS
		// ==========================================================================================

		private void SendSpeedToSlaves()
		{
			//
            Hid.USB_TxData[0] = (byte)Cmd.SPEED;
			Hid.USB_TxData[1] = (byte)CommSpeed;
			//
			Hid.USB_Write_Read();
            Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
			// --------------------------------------------- error ?
			if (Hid.USB_RxData[0] != 0) {
				Execution_Error("SendSpeedToSlaves");
			}
			else {
				Execution_OK();
			}
		}


		// ==========================================================================================
		//   SMOOTHING FUNCTIONS
		// ==========================================================================================

		internal void SmoothValue(ref float value, float new_value, float speed)
		{
			if (float.IsNaN(new_value)) {
				value = new_value;
				return;
			}

            // --------------------------------------------------- exclude gradually the smoothing 
            switch ((int)speed)
            {
                case 98: speed = 150; break;
                case 99: speed = 200; break;
                case 100: speed = 300; break;
            }

			float delta = new_value - value;
			float delta_sgn = Math.Sign(delta);
			float delta_abs = Math.Abs(delta);
			float delta_pow = 0;

			delta_pow = (float)speed / CommFps * delta_abs;

			if (delta_pow >= delta_abs) {
				value = new_value;
			}
			else {
				value += delta_sgn * delta_pow;
			}
		}

		internal void SmoothValue_Pow(ref float value, float new_value, float speed)
		{
			if (float.IsNaN(new_value)) {
				value = new_value;
				return;
			}

			float delta = new_value - value;
			float delta_sgn = Math.Sign(delta);
			float delta_abs = Math.Abs(delta);
			float delta_pow = 0;

			delta_pow = (float)Math.Pow((speed / CommFps * delta_abs), 2);

			if (delta_pow >= delta_abs) {
				value = new_value;
			}
			else {
				value += delta_sgn * delta_pow;
			}
		}


		// ==========================================================================================
		//   LOOPS AND COMMUNICATIONS
		// ==========================================================================================


        private System.Diagnostics.Stopwatch sw1 = new System.Diagnostics.Stopwatch();
		private void DataExchange()
		{
            //CommMillisec = (float)(sw1.Elapsed.TotalMilliseconds - 0.1);
            //CommMillisec = (float)(sw1.Elapsed.TotalMilliseconds);
			CommMillisec = (float)(sw1.ElapsedMilliseconds);
			sw1.Reset();
			sw1.Start();
			if (CommMillisec >= 1) {
				SmoothValue_Pow(ref CommFps, 1000 / CommMillisec, 5);
			}
			try {
				if (FastComm) {
					FastDataExchange();
				}
				else {
					SlowDataExchange();
				}
			}

			catch {
                	#if DEBUG
				    Interaction.Beep();
				    //MsgBox("Exception in ""DataExchange""")
				    Debug.Print("Exception in \"DataExchange\"");	
                    #endif
                  }
        }


		private void SlowDataExchange()
		{
			Int32 ix = default(Int32);
			byte[] b = null;
			bool ExchangedFlag = false;
			// -------------------------------------------------------- MMF to USB
			foreach (Slave s in Slaves) {
				Int32 nbytes = s.GetHostToMasterBytesCount();
				if (nbytes > 0) {
					ix = 0;
					b = new byte[nbytes];
					foreach (Pin p in s.Pins) {
						if (p.Direction == Pin.Directions.HostToMaster) {
							p.ReadValueFromMMF();
							p.ConvertValueToHardwareFormat();
							switch (p.GetPinType()) {
								case Pin.PinTypes.DIG_OUT:
									b[ix] = (byte)p.Value_RawInteger;
									ix += 1;
									break;
								case Pin.PinTypes.PWM_8:
								case Pin.PinTypes.SERVO_8:
									b[ix] = (byte)(p.Value_RawInteger / 256);
									ix += 1;
									break;
								case Pin.PinTypes.PWM_16:
								case Pin.PinTypes.SERVO_16:
                                    byte[] bytes = Module_Utils.ByteArrayFromUint16((ushort)p.Value_RawInteger);
									b[ix] = bytes[0];
									ix += 1;
									b[ix] = bytes[1];
									ix += 1;
									break;
							}
						}
					}
					SendValuesToSlave(s.mSlaveId, ref b);
					ExchangedFlag = true;
				}
			}
			// -------------------------------------------------------- USB to MMF
			foreach (Slave s in Slaves) {
				Int32 nbytes = s.GetMasterToHostBytesCount();
				if (nbytes > 0) {
					ExchangedFlag = true;
					ix = 0;
					b = new byte[nbytes];
					if (GetValuesFromSlave(s.mSlaveId, ref b)) {
						foreach (Pin p in s.Pins) {
							if (p.Direction == Pin.Directions.MasterToHost) {
								switch (p.UsbBytesCount) {
									case 1:
										p.Value_RawInteger = b[ix];
										ix += 1;
										break;
									case 2:
										p.Value_RawInteger = b[ix] * 256u + b[ix + 1];
										ix += 2;
										break;
									case 3:
										p.Value_RawInteger = b[ix] * 65536u + b[ix + 1] * 256u + b[ix + 2];
										ix += 3;
										break;
									case 4:
										p.Value_RawInteger = b[ix] * 256u * 65536u + b[ix + 1] * 65536u + b[ix + 2] * 256u + b[ix + 3];
										ix += 4;
										break;
								}
								p.ConvertValueFromHardwareFormat();
								p.WriteValueToMMF();
							}
						}
					}
				}
			}
			// --------------------------------------------------------  do not charge the cpu if not exchanged
			if (!ExchangedFlag) {
				Hid.USB_Write_Read();
			}
		}



		private void FastDataExchange()
		{
			// -------------------------------------------------------- fast data exchange
            Hid.USB_TxData[0] = (byte)Cmd.FAST_DATA_EXCHANGE;

			// -------------------------------------------------------- MMF to USB
			foreach (Slave s in Slaves) {
				foreach (Pin p in s.Pins) {
					if (p.Direction == Pin.Directions.HostToMaster) {
						p.ReadValueFromMMF();
						p.ConvertValueToHardwareFormat();
						p.WriteValueToUsbBuffer();
					}
				}
			}

			// -------------------------------------------------------- USB <> Master <> Slaves
			//Hid.disable_for_test = True
			Hid.USB_Write_Read();
			//Hid.disable_for_test = False
            Hid.USB_TxData[0] = (byte)Cmd.NOCMD;

			// -------------------------------------------------------- manage errors
			if (Hid.USB_RxData[0] == 0) {
				// ---------------------------------------------------- USB to MMF
				foreach (Slave s in Slaves) {
					foreach (Pin p in s.Pins) {
						if (p.Direction == Pin.Directions.MasterToHost) {
							p.ReadValueFromUsbBuffer();
							p.ConvertValueFromHardwareFormat();
							p.WriteValueToMMF();
						}
					}
				}
				Execution_OK();
			}
			else {
				Execution_Error("FastDataExchange");
			}

		}

		internal bool GetValuesFromSlave(Int32 SlaveId, ref byte[] ByteArray)
		{
			// ----------------------------------------------------
			// SyncLock not needed because called from timer
			// ---------------------------------------------------- send CMD / SlaveId / Nbytes
            Hid.USB_TxData[0] = (byte)Cmd.GET_VALUES;
			Hid.USB_TxData[1] = (byte)SlaveId;
			Hid.USB_TxData[2] = (byte)ByteArray.Length;
			// ---------------------------------------------------- send and receive
			Hid.USB_Write_Read();
            Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
			// ---------------------------------------------------- get bytes
			for (Int32 i = 0; i <= ByteArray.Length - 1; i++) {
				ByteArray[i] = Hid.USB_RxData[i + 1];
			}
			// ---------------------------------------------------- manage errors
			if (Hid.USB_RxData[0] != 0) {
				Execution_Error("GetValuesFromSlave");
				return false;
			}
			else {
				Execution_OK();
			}
			return true;
		}

		internal void SendValuesToSlave(Int32 SlaveId, ref byte[] bytes)
		{
			// ----------------------------------------------------
			// SyncLock not needed because called from timer
			// ---------------------------------------------------- send CMD / SlaveId / Nbytes
            Hid.USB_TxData[0] = (byte)Cmd.SEND_VALUES;
			Hid.USB_TxData[1] = (byte)SlaveId;
			Hid.USB_TxData[2] = (byte)bytes.Length;
			// ---------------------------------------------------- send bytes
			for (Int32 i = 0; i <= bytes.Length - 1; i++) {
				Hid.USB_TxData[3 + i] = bytes[i];
			}
			// ---------------------------------------------------- send and receive
			Hid.USB_Write_Read();
            Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
			// ---------------------------------------------------- manage errors
			if (Hid.USB_RxData[0] != 0) {
				Execution_Error("SendValuesToSlave");
			}
			else {
				Execution_OK();
			}
		}



		internal void GetBytesFromSlave(Int32 SlaveId, ref byte[] ByteArray)
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				// ---------------------------------------------------- send CMD / SlaveId / Nbytes
                Hid.USB_TxData[0] = (byte)Cmd.GET_BYTES;
				Hid.USB_TxData[1] = (byte)SlaveId;
				Hid.USB_TxData[2] = (byte)ByteArray.Length;
				// ---------------------------------------------------- send and receive
				Hid.USB_Write_Read();
                Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
				// ---------------------------------------------------- get bytes
				for (Int32 i = 0; i <= ByteArray.Length - 1; i++) {
					ByteArray[i] = Hid.USB_RxData[i + 1];
				}
				// ---------------------------------------------------- manage errors
				if (Hid.USB_RxData[0] != 0) {
					Execution_Error("GetBytesFromSlave");
				}
				else {
					Execution_OK();
				}
			}
			DataExchangeThread_Start();
		}

		internal void SendBytesToSlave(Int32 SlaveId, ref byte[] bytes)
		{
			DataExchangeThread_Stop();
			lock (ActionLock) {
				// ---------------------------------------------------- send CMD / SlaveId / Nbytes
                Hid.USB_TxData[0] = (byte)Cmd.SEND_BYTES;
				Hid.USB_TxData[1] = (byte)SlaveId;
				Hid.USB_TxData[2] = (byte)bytes.Length;
				// ---------------------------------------------------- send bytes
				for (Int32 i = 0; i <= bytes.Length - 1; i++) {
					Hid.USB_TxData[3 + i] = bytes[i];
				}
				// ---------------------------------------------------- send and receive
				Hid.USB_Write_Read();
                Hid.USB_TxData[0] = (byte)Cmd.NOCMD;
				// ---------------------------------------------------- manage errors
				if (Hid.USB_RxData[0] != 0) {
					Execution_Error("SendBytesToSlave");
				}
				else {
					Execution_OK();
				}
			}
			DataExchangeThread_Start();
		}


		private void Execution_Error(string ErrorString)
		{
			ErrorRate += 0.01f;
			Debug.Print("ERROR - " + ErrorString);
		}

		private void Execution_OK()
		{
			ErrorRate *= 0.95f;
		}


	}
}

