using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;

using System.Globalization;
//using System.Math;
namespace Theremino_HAL
{



	// ==================================================================================================
	//   CLASS SLAVE
	// ==================================================================================================

	class Slave
	{
		public enum SlaveTypes : int
		{
			Custom = 0,
			CapSensor = 1,
			Servo = 2,
			Generic = 3,
			InOut = 4,
			MasterPins = 5,
			RmsLogger = 6,
			Sniffer = 7,
			Unknown = 255
		}

		internal Int32 MasterId;
		internal Int32 mSlaveId;
		internal SlaveTypes SlaveType;
		internal List<Pin> Pins;

		internal Int32 LoadCurrent_mA;

		internal Int32 SlaveId {
			get { return mSlaveId; }
// test if mSlaveId ok
			set { mSlaveId = value; }
		}

		internal string SlaveTypeString()
		{
			switch (SlaveType) {
				case SlaveTypes.Custom:
					return "Custom";
				case SlaveTypes.CapSensor:
					return "CapSensor";
				case SlaveTypes.Servo:
					return "Servo";
				case SlaveTypes.Generic:
					return "Generic";
				case SlaveTypes.InOut:
					return "InOut";
				case SlaveTypes.MasterPins:
					return "MasterPins";
				case SlaveTypes.RmsLogger:
					return "RmsLogger";
				case SlaveTypes.Sniffer:
					return "Sniffer";
				case SlaveTypes.Unknown:
					return "Unknown";
				default:
					return "Unknown";
			}
		}


		internal Slave(SlaveTypes _slavetype, Int32 _masterid, Int32 _slaveid)
		{
			SlaveType = _slavetype;
			MasterId = _masterid;
			SlaveId = _slaveid;
			LoadCurrent_mA = 0;
		}

		internal Int32 GetHostToMasterBytesCount()
		{
			Int32 functionReturnValue = default(Int32);
			functionReturnValue = 0;
			foreach (Pin p in Pins) {
				if (p.Direction == Pin.Directions.HostToMaster) {
					functionReturnValue += p.UsbBytesCount;
				}
			}
			return functionReturnValue;
		}

		internal Int32 GetMasterToHostBytesCount()
		{
			Int32 functionReturnValue = default(Int32);
			functionReturnValue = 0;
			foreach (Pin p in Pins) {
				if (p.Direction == Pin.Directions.MasterToHost) {
					functionReturnValue += p.UsbBytesCount;
				}
			}
			return functionReturnValue;
		}


        internal bool CalibrationNeeded()
        {
            foreach (Pin p in Pins)
            {
                switch (p.GetPinType())
                {
                    case Pin.PinTypes.CAP_SENSOR: // Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                        return true;
                }
            }
            return false;
        }

		internal virtual void Calibrate()
		{
			foreach (Pin p in Pins) {
				p.Calibrate();
			}
		}

		internal virtual void CalibrateRawZero()
		{
			foreach (Pin p in Pins) {
				p.CalibrateRawZero();
			}
		}

		internal virtual UInt32 GetCapSensor_Value_RawInteger_Delta()
		{
			return 0;
		}


		internal virtual void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (SlaveType) {
				case SlaveTypes.Servo:
				case SlaveTypes.Generic:
				case SlaveTypes.InOut:

					switch (pinType) {

						case Pin.PinTypes.UNUSED:
						case Pin.PinTypes.DIG_OUT:
						case Pin.PinTypes.PWM_8:
						case Pin.PinTypes.PWM_16:
						case Pin.PinTypes.SERVO_8:
						case Pin.PinTypes.SERVO_16:
						case Pin.PinTypes.DIG_IN:
						case Pin.PinTypes.DIG_IN_PU:
						case Pin.PinTypes.COUNTER:
						case Pin.PinTypes.COUNTER_PU:
							Pins[npin].SetPinType(pinType);

							break;
						case Pin.PinTypes.ADC_8:
						case Pin.PinTypes.ADC_16:
						case Pin.PinTypes.CAP_8:
						case Pin.PinTypes.CAP_16:
						case Pin.PinTypes.RES_8:
						case Pin.PinTypes.RES_16:
							if (npin < 8) Pins[npin].SetPinType(pinType); 

							break;
						case Pin.PinTypes.FAST_COUNTER:
						case Pin.PinTypes.FAST_COUNTER_PU:
							if (npin == 7) Pins[npin].SetPinType(pinType); 

							break;
						case Pin.PinTypes.USOUND_SENSOR:
						case Pin.PinTypes.PERIOD:
						case Pin.PinTypes.PERIOD_PU:
							if (npin == 8) Pins[npin].SetPinType(pinType); 

							break;
						// TODO add MCA_8_BYTES, MCA_16_BYTES, MCA_32_BYTES
					}

					break;
			}
		}

		internal virtual void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			switch (SlaveType) {

				case SlaveTypes.Servo:
				case SlaveTypes.Generic:
				case SlaveTypes.InOut:
					cmb.Items.Clear();
					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16));
					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU));
					//
					if (p.PinId < 8) {
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16));
					}

					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU));

					if (p.PinId == 7) {
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU));
					}


					if (p.PinId == 8) {
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR));
					}


					break;
				// TODO add MCA_8_BYTES, MCA_16_BYTES, MCA_32_BYTES

				default:
					cmb.Items.Clear();
					cmb.Items.Add(Pin.PinTypeToString(p.GetPinType()));

					break;
			}
		}

	}
}
namespace Theremino_HAL
{





	// ==================================================================================================
	// SUBCLASS - Slave_CapSensor
	// ==================================================================================================

	class Slave_CapSensor : Slave
	{

		internal Slave_CapSensor(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.CapSensor, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 12;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
		}

		internal override UInt32 GetCapSensor_Value_RawInteger_Delta()
		{
			UInt32 functionReturnValue = default(UInt32);
			{
                UInt32 n = Module_MathFunctions.AbsUint32Difference(Pins[0].Value_RawInteger_Max, Pins[0].Value_RawInteger_Min);
				if (n < 0) n = 0; 
				if (n > 400000) n = 0; 
				functionReturnValue = n;
				Pins[0].ResetPosMinMax();
			}
			return functionReturnValue;
		}

		internal override void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (pinType) {
				case Pin.PinTypes.UNUSED:
				case Pin.PinTypes.CAP_SENSOR:
					Pins[npin].SetPinType(pinType);
					break;
			}
		}

		internal override void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			cmb.Items.Clear();
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_SENSOR));
		}

	}
}
namespace Theremino_HAL
{


	// ==================================================================================================
	//   SUBCLASS - Slave_Servo
	// ==================================================================================================

	class Slave_Servo : Slave
	{

		internal Slave_Servo(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.Servo, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 80;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
		}

	}
}
namespace Theremino_HAL
{



	// ==================================================================================================
	//   SUBCLASS - Slave_Generic
	// ==================================================================================================

	class Slave_Generic : Slave
	{

		internal Slave_Generic(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.Generic, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 80;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11));
		}

	}
}
namespace Theremino_HAL
{



	// ==================================================================================================
	//   SUBCLASS - Slave_InOut
	// ==================================================================================================

	class Slave_InOut : Slave
	{

		internal Slave_InOut(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.InOut, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 80;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11));
		}

	}
}
namespace Theremino_HAL
{



	// ==================================================================================================
	//   SUBCLASS - Slave_MasterPins
	// ==================================================================================================

	class Slave_MasterPins : Slave
	{

		internal Slave_MasterPins(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.MasterPins, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 10;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
		}

		internal override void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (pinType) {

				case Pin.PinTypes.UNUSED:
				case Pin.PinTypes.DIG_OUT:
				case Pin.PinTypes.PWM_8:
				case Pin.PinTypes.PWM_16:
				case Pin.PinTypes.SERVO_8:
				case Pin.PinTypes.SERVO_16:
				case Pin.PinTypes.DIG_IN:
				case Pin.PinTypes.DIG_IN_PU:
				case Pin.PinTypes.ADC_8:
				case Pin.PinTypes.ADC_16:
				case Pin.PinTypes.CAP_8:
				case Pin.PinTypes.CAP_16:
				case Pin.PinTypes.RES_8:
				case Pin.PinTypes.RES_16:
				case Pin.PinTypes.COUNTER:
				case Pin.PinTypes.COUNTER_PU:
					Pins[npin].SetPinType(pinType);

					break;
				case Pin.PinTypes.FAST_COUNTER:
				case Pin.PinTypes.FAST_COUNTER_PU:
					if (!AlreadyUsed_FastCounter(Pins[npin])) Pins[npin].SetPinType(pinType); 

					break;
				case Pin.PinTypes.USOUND_SENSOR:
				case Pin.PinTypes.PERIOD:
				case Pin.PinTypes.PERIOD_PU:
					if (!AlreadyUsed_PeriodCounter(Pins[npin])) Pins[npin].SetPinType(pinType); 

					break;
				// TODO add MCA_8_BYTES, MCA_16_BYTES, MCA_32_BYTES
			}
		}

		internal override void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			cmb.Items.Clear();
			//
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
			//
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16));
			//
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU));

			if (!AlreadyUsed_FastCounter(p)) {
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER));
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU));
			}

			if (!AlreadyUsed_PeriodCounter(p)) {
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD));
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU));
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR));
			}

			// TODO add MCA_8_BYTES, MCA_16_BYTES, MCA_32_BYTES
		}

		private bool AlreadyUsed_FastCounter(Pin pinSelected)
		{
			foreach (Pin p in Pins) {
				if (object.ReferenceEquals(p, pinSelected)) continue; 
				if (p.GetPinType() == Pin.PinTypes.FAST_COUNTER || p.GetPinType() == Pin.PinTypes.FAST_COUNTER_PU) {
					return true;
				}
			}
			return false;
		}

		private bool AlreadyUsed_PeriodCounter(Pin pinSelected)
		{
			foreach (Pin p in Pins) {
				if (object.ReferenceEquals(p, pinSelected)) continue; 
				if (p.GetPinType() == Pin.PinTypes.PERIOD || p.GetPinType() == Pin.PinTypes.PERIOD_PU || p.GetPinType() == Pin.PinTypes.USOUND_SENSOR) {
					return true;
				}
			}
			return false;
		}
	}
}
namespace Theremino_HAL
{



	// ==================================================================================================
	//   SUBCLASS - Slave_RmsLogger
	// ==================================================================================================

	class Slave_RmsLogger : Slave
	{

		internal Slave_RmsLogger(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.RmsLogger, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 10;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.ADC_16, MasterId, mSlaveId, 0));
			// ADC - Battery voltage 
		}

	}
}
namespace Theremino_HAL
{


	// ==================================================================================================
	//   SUBCLASS - Slave_Sniffer
	// ==================================================================================================

	class Slave_Sniffer : Slave
	{

		internal Slave_Sniffer(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.Sniffer, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 10;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.SNIFFER, MasterId, mSlaveId, 0));
		}

	}
}



