﻿
using Microsoft.VisualBasic;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Windows.Forms;

namespace Theremino_HAL
{
    public partial class Form2 : Form
    {
        private void Form2_Load(object sender, System.EventArgs e)
        {
            btn_TextLog.ClickButtonArea += btn_GraphicLog_ClickButtonArea;
            btn_SetRawZero.ClickButtonArea += btn_SetRawZero_ClickButtonArea;
            btn_ShowRawCount.ClickButtonArea += btn_ShowRawCount_ClickButtonArea;

            rtb_SnifferLog.Location = pbox_ScrollingScope.Location;
            rtb_SnifferLog.Size = pbox_ScrollingScope.Size;

            SetScaleLabels();
        }

        private void Form2_VisibleChanged(object sender, System.EventArgs e)
        {
            StartTimer();
        }

        internal void StopTimer()
        {
            Timer_100Hz.Stop();
        }

        internal void StartTimer()
        {
            if (Visible & WindowState != FormWindowState.Minimized)
            {
                Timer_100Hz.Interval = 10;
                Timer_100Hz.Start();
            }
            else
            {
                Timer_100Hz.Stop();
            }
        }

        private void Form2_FormClosing(object sender, System.Windows.Forms.FormClosingEventArgs e)
        {
            GroupBox1.BackColor = Color.LightGoldenrodYellow;
            ThereminoSystem.TheSystem_StopAllSnifferLogs();
            this.Hide();
            e.Cancel = true;
        }

        private void Form2_LocationChanged(object sender, System.EventArgs e)
        {
            if (!Module_SaveLoad.EventsAreEnabled) return;

            Module_SaveLoad.LimitFormPosition(this);
        }


        private void Timer_100Hz_Tick(System.Object sender, System.EventArgs e)
        {
            ShowPinDetails();
        }


        private Pin AlternatePin;

        private Pin SelectedPin;
        internal void DeselectAllPins()
        {
            SelectedPin = null;
            AlternatePin = null;
            rtb_SnifferLog.Visible = false;
            pbox_ScrollingScope.Visible = true;
            ThereminoSystem.TheSystem_StopAllSnifferLogs();
            LabelPin1.Text = "- - -";
            lbl_Details1.Text = "";
            Module_FillBars.FillPictureBox(pbox_Details1, 0, Color.Orange, Color.AliceBlue, 0);
            LabelPin2.Text = "- - -";
            lbl_Details2.Text = "";
            Module_FillBars.FillPictureBox(pbox_Details2, 0, Color.Orange, Color.AliceBlue, 0);
        }


        internal void SetPinDetails()
        {
            Pin p = ThereminoSystem.GetSelectedPin();
            //
            if (p == null) return;
            //
            if (object.ReferenceEquals(p, SelectedPin))
            {
                AlternatePin = null;
                SelectedPin = p;
            }
            else
            {
                AlternatePin = SelectedPin;
                SelectedPin = p;
            }
            //
            if (SelectedPin != null)
            {
                LabelPin1.Text = ComposePinString(SelectedPin);
            }
            if (AlternatePin != null)
            {
                LabelPin2.Text = ComposePinString(AlternatePin);
            }
            else
            {
                LabelPin2.Text = "- - -";
                lbl_Details2.Text = "";
                Module_FillBars.FillPictureBox(pbox_Details2, 0, Color.Orange, Color.AliceBlue, 0);
            }
            // ---------------------------------------------------------------------- visualize sniffer log
            if (SelectedPin != null && SelectedPin.GetPinType() == Pin.PinTypes.SNIFFER && this.Visible)
            {
                if (!SelectedPin.SnifferLogEnabled)
                {
                    if (btn_TextLog.Checked)
                    {
                        rtb_SnifferLog.Visible = true;
                        pbox_ScrollingScope.Visible = false;
                        GroupBox1.BackColor = Color.Tomato;
                        ThereminoSystem.TheSystem_StopAllSnifferLogs();
                        rtb_SnifferLog.Text = "";
                        SelectedPin.SnifferLogEnabled = true;
                    }
                }
                else
                {
                    if (!btn_TextLog.Checked)
                    {
                        rtb_SnifferLog.Visible = false;
                        pbox_ScrollingScope.Visible = true;
                    }
                    GroupBox1.BackColor = Color.LightGoldenrodYellow;
                    ThereminoSystem.TheSystem_StopAllSnifferLogs();
                }
            }
            //
            ThereminoSystem.TheSystem_CalibrateRawZero();
            SetScaleLabels();
        }

        private void btn_GraphicLog_ClickButtonArea(System.Object Sender, System.EventArgs e)
        {
            if (!btn_TextLog.Checked)
            {
                rtb_SnifferLog.Visible = false;
                pbox_ScrollingScope.Visible = true;
                GroupBox1.BackColor = Color.LightGoldenrodYellow;
                ThereminoSystem.TheSystem_StopAllSnifferLogs();
            }
            else
            {
                rtb_SnifferLog.Visible = true;
                pbox_ScrollingScope.Visible = false;
            }
        }

        private void ShowPinDetails()
        {
            ShowDetails();
            UpdateScrollingScope();
            BeepOnChange();
        }

        private string ComposePinString(Pin p)
        {
            return "M:" + (p.MasterId + 1).ToString() + " S:" + (p.SlaveId + 1).ToString() + " Pin:" + (p.PinId + 1).ToString();
        }

        private void ShowDetails()
        {
            if (SelectedPin != null)
            {
                lbl_Details1.Text = SelectedPin.GetValueString();
                Module_FillBars.FillPictureBox(pbox_Details1, SelectedPin.Value_Normalized, Color.Orange, Color.AliceBlue, 0);
            }
            if (AlternatePin != null)
            {
                lbl_Details2.Text = AlternatePin.GetValueString();
                Module_FillBars.FillPictureBox(pbox_Details2, AlternatePin.Value_Normalized, Color.Orange, Color.AliceBlue, 0);
            }
        }

        private void BeepOnChange()
        {
            if (!btn_BeepOnChange.Checked) return;
            if (SelectedPin != null)
            {
                double c = SelectedPin.Value;
                if (c != static_BeepOnChange_oldc & c != 0)
                {
                    //Debug.Print(SelectedPin.Value.ToString());
                    Interaction.Beep();
                    static_BeepOnChange_oldc = c;
                }
            }
        }
        static double static_BeepOnChange_oldc = 0;

        private void btn_SetRawZero_ClickButtonArea(System.Object Sender, System.EventArgs e)
        {
            ThereminoSystem.TheSystem_CalibrateRawZero();
            SetScaleLabels();
        }

        private void btn_ShowRawCount_ClickButtonArea(System.Object Sender, System.EventArgs e)
        {
            ThereminoSystem.TheSystem_CalibrateRawZero();

            if (btn_ShowRawCount.Checked)
            {
                tkbar_VerticalScaleZoom.Value = 100;
            }
            else
            {
                tkbar_VerticalScaleZoom.Value = 1;
            }
            SetScaleLabels();
        }

        private void UpdateScrollingScope()
        {
            if (!pbox_ScrollingScope.Visible) return;
            Single v1 = -999;
            Single v2 = -999;
            Single zoom = tkbar_VerticalScaleZoom.Value;
            if (btn_ShowRawCount.Checked)
            {
                if (SelectedPin != null)
                {
                    v1 = (SelectedPin.Value_RawInteger_Zero - (int)SelectedPin.Value_RawInteger) * zoom / 100.0F;
                }
                if (AlternatePin != null)
                {
                    v2 = (AlternatePin.Value_RawInteger_Zero - (int)AlternatePin.Value_RawInteger) * zoom / 100.0F;
                }
            }
            else
            {
                if (SelectedPin != null)
                {
                    v1 = pbox_ScrollingScope.Height * 0.49F * (0.5F - SelectedPin.Value_Normalized) * zoom;
                }
                if (AlternatePin != null)
                {
                    v2 = pbox_ScrollingScope.Height * 0.49F * (0.5F - AlternatePin.Value_Normalized) * zoom;
                }
            }
            Module_FillBars.DisplayCounterscope(pbox_ScrollingScope, v1, v2, tkbar_ScrollSpeed.Value);
        }

         private void SetScaleLabels()
         {
            Label_Scale3.Text = "";
            Label_Scale2.Text = "";
            Label_Scale1.Text = "";
            Label_Scale3b.Text = "";
            Label_Scale2b.Text = "";
            Label_Scale1b.Text = "";
            double zoom  = tkbar_VerticalScaleZoom.Value;
            if (btn_ShowRawCount.Checked)
            {
                if (SelectedPin != null)
                {
                    Label_Scale3.Text = (SelectedPin.Value_RawInteger_Zero + 6800 / zoom).ToString("0");
                    Label_Scale2.Text = SelectedPin.Value_RawInteger_Zero.ToString("0");
                    Label_Scale1.Text = (SelectedPin.Value_RawInteger_Zero - 6800 / zoom).ToString("0");
                }
                if (AlternatePin != null)
                {
                    Label_Scale3b.Text = (AlternatePin.Value_RawInteger_Zero + 6800 / zoom).ToString("0");
                    Label_Scale2b.Text = AlternatePin.Value_RawInteger_Zero.ToString("0");
                    Label_Scale1b.Text = (AlternatePin.Value_RawInteger_Zero - 6800 / zoom).ToString("0");
                }
             }
             else
             {
                Label_Scale3.Text = (500 + 500 / zoom).ToString("0");
                Label_Scale2.Text = "500";
                Label_Scale1.Text = (500 - 500 / zoom).ToString("0");
             }
         }

         private void tkbar_VerticalScaleZoom_Scroll(object sender, EventArgs e)
         {
             SetScaleLabels();
         }




        public Form2()
        {
            LocationChanged += Form2_LocationChanged;
            FormClosing += Form2_FormClosing;
            SizeChanged += Form2_VisibleChanged;
            VisibleChanged += Form2_VisibleChanged;
            Load += Form2_Load;
            InitializeComponent();
        }
    }
}
