using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
using System.Reflection;
namespace Theremino_HAL
{

	public static class Module_SaveLoad
	{
        /// <summary> 
        /// Gets the name of the application. 
        /// </summary> 
        /// <value></value> 
        /// <remarks></remarks> 
        public static string ApplicationName
        {
            get
            {
                var entryAssembly = Assembly.GetEntryAssembly();
                var applicationTitle = ((AssemblyTitleAttribute)entryAssembly.GetCustomAttributes(typeof(AssemblyTitleAttribute), false)[0]).Title;
                if (string.IsNullOrEmpty(applicationTitle))
                {
                    applicationTitle = entryAssembly.GetName().Name;
                }
                return applicationTitle;
            }
        } 

		public static bool EventsAreEnabled = false;
		//Friend Form1_VisibleAtStart As Boolean

		public static bool Form2_VisibleAtStart;

		// =======================================================================================================
		//   APP TITLE AND VERSION
		// =======================================================================================================
        public static string AppTitleAndVersion([System.Runtime.InteropServices.OptionalAttribute, System.Runtime.InteropServices.DefaultParameterValueAttribute("")]  // ERROR: Optional parameters aren't supported in C#
string Title)
		{
            if (string.IsNullOrEmpty(Title)) Title = ApplicationName.Replace('_', ' '); 
			string[] s = Theremino_HAL.My.MyProject.Application.Info.Version.ToString().Split('.');
            return Title + " - V" + s[0] + "." + s[1];
		}

        //  =======================================================================================
        //    FILES
        //  =======================================================================================
        public static bool FileExists(string fileName)
        {
            return System.IO.File.Exists(fileName);
        }


		// =======================================================================================================
		//   Platform selection
		// =======================================================================================================
		public static bool CanUseWindowsDlls()
		{
			switch (Environment.OSVersion.Platform) {
				case PlatformID.Win32NT:
				case PlatformID.Win32S:
				case PlatformID.Win32Windows:
				case PlatformID.WinCE:
					return true;
				default:
					return false;
			}
		}
        public static string PlatformAdjustedFileName(string FileName, string DefaultName)
		{
			if (CanUseWindowsDlls()) {
				FileName = FileName.Replace("/", "\\");
				return FileName;
			}
			else {
				if (FileName.Contains(":")) FileName = DefaultName; 
				FileName = FileName.Replace("\\", "/");
				return FileName;
			}
		}


		// =======================================================================================
		//  FORM
		// =======================================================================================
        public static void LimitFormPosition(System.Windows.Forms.Form f)
		{
            if (f == null) return;
			if (f.WindowState != FormWindowState.Normal) return; 
			GetMaxScreenBounds();
			EnsureFormVisible(f);
		}

		private static int sw;
		private static int sh;
		private static void GetMaxScreenBounds()
		{
			// -------------------------------------- Gets an array of all the screens connected to the system.
			System.Windows.Forms.Screen[] Screens = System.Windows.Forms.Screen.AllScreens;
			// -------------------------------------- For each screen
			int Index = 0;
			for (Index = 0; Index <= Screens.GetUpperBound(0); Index++) {
				sw = Math.Max(sw, Screens[Index].Bounds.Left + Screens[Index].Bounds.Width);
				sh = Math.Max(sh, Screens[Index].Bounds.Top + Screens[Index].Bounds.Height);
			}
		}

		//Private Sub EnsureFormCompletelyVisible(ByVal frm As Form)
		//    With frm
		//        .Width = Math.Min(.Width, sw)           ' non oltre le dimensioni di una maximized window
		//        .Height = Math.Min(.Height, sh)         ' non oltre le dimensioni di una maximized window
		//        .Width = Math.Max(.Width, 320)          ' non meno di 320x240
		//        .Height = Math.Max(.Height, 240)        ' non meno di 320x240
		//        .Left = Math.Min(.Left, sw - .Width)    ' non oltre il bordo destro
		//        .Top = Math.Min(.Top, sh - .Height)     ' non oltre il bordo inferiore
		//        .Left = Math.Max(.Left, 0)              ' non meno del bordo sinistro
		//        .Top = Math.Max(.Top, 0)                ' non sopra al bordo superiore
		//    End With
		//End Sub

		private static void EnsureFormVisible(Form frm)
		{
			{
				frm.Width = Math.Min(frm.Width, sw);
				// not more than the VIRTUALSCREEN dimensions
				frm.Height = Math.Min(frm.Height, sh);
				// not more than the VIRTUALSCREEN dimensions 
				frm.Width = Math.Max(frm.Width, 32);
				// at least 32x24
				frm.Height = Math.Max(frm.Height, 24);
				// at least 32x24
				frm.Left = Math.Min(frm.Left, sw - 50);
				// not beyond the right border - 50 pixels
				frm.Top = Math.Min(frm.Top, sh - 50);
				// not beyond the bottom border - 50 pixels
				frm.Left = Math.Max(frm.Left, 50 - frm.Width);
				// at least at the left border + 50 pixels
				frm.Top = Math.Max(frm.Top, -30);
				// at least at the top border
			}
		}

		// (The value of the RestoreBounds property is valid only 
		//   when the WindowState property of the Form class is not equal to Normal)
		public static Rectangle GetFormRectangle(Form frm)
		{
			Rectangle r = default(Rectangle);
			if (frm.WindowState == FormWindowState.Normal) {
				r = frm.Bounds;
			}
			else {
				r = frm.RestoreBounds;
			}
			return r;
		}


		// ================================================================================================
		//  Private Read-Write functions
		// ================================================================================================


		private static string TabString(string Name, double Value, string fmt)
		{
			Int32 nTab = Math.Max(0, 22 - Name.Length);
			if (double.IsNaN(Value)) {
				return Name;
			}
			else {
				return Name + "=" + Strings.StrDup(nTab, " ") + Value.ToString(fmt);
			}
		}


		private static string TabString(string Name, bool Value)
		{
			Int32 nTab = Math.Max(0, 22 - Name.Length);

			return Name + "=" + Strings.StrDup(nTab, " ") + Value.ToString();
		}


		private static string TabString(string Name, string Value)
		{
			Int32 nTab = Math.Max(0, 22 - Name.Length);

			return Name + "=" + Strings.StrDup(nTab, " ") + Value;
		}

		private static float Val_Single(string l)
		{
			return (float)Conversion.Val(l.Replace(",", "."));
		}

		private static double Val_Double(string l)
		{
			return Conversion.Val(l.Replace(",", "."));
		}

		private static Int32 Val_Int(string l)
		{
			return (int)Conversion.Val(l);
		}

		public static string ExtractParamName(ref string s)
		{
			string functionReturnValue = null;
			// ------------------------- Returns the first field from begin to the first "=" symbol
			// -------------------------  and removes it from the string
			int i = 0;
			i = Strings.InStr(s, "=", Microsoft.VisualBasic.CompareMethod.Binary);
			if (i > 0) {
				functionReturnValue = Strings.Trim(Strings.Left(s, i - 1));
				s = Strings.Trim(Strings.Mid(s, i + 1));
			}
			else {
				functionReturnValue = Strings.Trim(s);
				s = "";
			}
			return functionReturnValue;
		}

		private static string AssemblyName()
		{
			return System.Reflection.Assembly.GetExecutingAssembly().GetName().Name;
		}



		// ==================================================================================================
		//  SAVE LOAD -- Program INI
		// ==================================================================================================

        public static void Save_INI()
		{
			string iniFileName = Module_SaveLoad.PlatformAdjustedFileName(Application.StartupPath + "\\" + AssemblyName() + "_INI.txt", "");
			System.IO.StreamWriter f = null;
            f = System.IO.File.CreateText(iniFileName);
			//
			f.WriteLine("");
			f.WriteLine("===========================================");
			f.WriteLine(" Program Params");
			f.WriteLine("===========================================");
			// ------------------------------------------------------------------------------ FORM BOUNDS
			Rectangle r = default(Rectangle);
            r = GetFormRectangle(My.MyProject.Forms.Form1);
			f.WriteLine(TabString("Form1_Top", r.Top, ""));
			f.WriteLine(TabString("Form1_Left", r.Left, ""));
			f.WriteLine(TabString("Form1_Width", r.Width, ""));
			f.WriteLine(TabString("Form1_Height", r.Height, ""));
			f.WriteLine(TabString("Form1_WindowState", (int)My.MyProject.Forms.Form1.WindowState, ""));
            //f.WriteLine(TabString("Form1_VisibleAtStart", My.MyProject.Forms.Form1.Visible And My.MyProject.Forms.Form1.WindowState <> FormWindowState.Minimized))

            r = GetFormRectangle(My.MyProject.Forms.Form2);
			f.WriteLine(TabString("Form2_Top", r.Top, ""));
			f.WriteLine(TabString("Form2_Left", r.Left, ""));
			f.WriteLine(TabString("Form2_VisibleAtStart", My.MyProject.Forms.Form2.Visible & My.MyProject.Forms.Form2.WindowState != FormWindowState.Minimized));
			//
			f.WriteLine("");
			f.WriteLine(TabString("MinChange", My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger, ""));
            f.WriteLine(TabString("BeepOnErrors", My.MyProject.Forms.Form1.ToolStripButton_BeepOnErrors.Checked));
            f.WriteLine(TabString("Language", Utils_LocaleNames.Language));
			//
			f.Close();
		}



        public static void Load_INI()
		{
            // ERROR: Not supported in C#: ReDimStatement
            ThereminoSystem.ConfigDatabase = new object[0]; 
			

			string[] Config = new string[-1 + 1];
			// ------------------------------------------------------------- defaults
			//Form1_VisibleAtStart = True
			My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger = 30;
			// -------------------------------------------------------------
			string l = null;
			string iniFileName = PlatformAdjustedFileName(Application.StartupPath + "\\" + AssemblyName() + "_INI.txt", "");
			if (Theremino_HAL.My.MyProject.Computer.FileSystem.FileExists(iniFileName)) {
				//
				System.IO.StreamReader f = null;
				f = System.IO.File.OpenText(iniFileName);
				//
				while (!f.EndOfStream) {
					l = f.ReadLine();
					switch (ExtractParamName(ref l)) {
						case "Form1_Top":
							My.MyProject.Forms.Form1.Top = Val_Int(l);
							break;
						case "Form1_Left":
							My.MyProject.Forms.Form1.Left = Val_Int(l);
							break;
						case "Form1_Width":
							My.MyProject.Forms.Form1.Width = Val_Int(l);
							break;
						case "Form1_Height":
							My.MyProject.Forms.Form1.Height = Val_Int(l);
							break;
						case "Form1_WindowState":
							My.MyProject.Forms.Form1.WindowState = (FormWindowState)(Conversion.Val(l));
							My.MyProject.Forms.Form1.Refresh();
							break;
						//Case "Form1_VisibleAtStart" : Form1_VisibleAtStart = l = "True"
						//
						case "Form2_Top":
							My.MyProject.Forms.Form2.Top = Val_Int(l);
							break;
						case "Form2_Left":
							My.MyProject.Forms.Form2.Left = Val_Int(l);
							break;
						case "Form2_VisibleAtStart":
							Form2_VisibleAtStart = l == "True";
							break;
						//
						case "MinChange":
							My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger = Val_Int(l);
							break;
						case "BeepOnErrors":
                            My.MyProject.Forms.Form1.ToolStripButton_BeepOnErrors.Checked = l == "True";
							break;
                        case "Language": 
                            Utils_LocaleNames.Language = l;
                            break;
					}
				}
				f.Close();
			}
			//
			{
				if (My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger < 0) My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger = 0; 
				if (My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger > 100) My.MyProject.Forms.Form1.txt_MinChange.NumericValueInteger = 100; 
			}
			//
            LimitFormPosition(My.MyProject.Forms.Form1);
            LimitFormPosition(My.MyProject.Forms.Form2);
		}




		// ==================================================================================================
		//  SAVE LOAD -- Configurations
		// ==================================================================================================

        public static void Save_ConfigDatabase()
		{
			//
			string iniFileName = PlatformAdjustedFileName(Application.StartupPath + "\\" + AssemblyName() + "_ConfigDatabase.txt", "");
			System.IO.StreamWriter f = null;
            f = System.IO.File.CreateText(iniFileName);
			//
            for (Int32 c = 0; c <= ThereminoSystem.ConfigDatabase.Length - 1; c++)
            {
                string[] config = ThereminoSystem.ConfigDatabase[c] as string[];
				AlignConfigurationSpacing(ref config);
				f.WriteLine("");
				f.WriteLine("Component  Name                Param1    Param2    Param3    Param4    Param5    Param6    Param7");
				f.WriteLine("-------------------------------------------------------------------------------------------------");
				for (Int32 i = 0; i <= config.Length - 1; i++) {
					f.WriteLine(config[i]);
				}
				f.WriteLine("");
			}
			//
			f.Close();
		}

		public static Int32[] SpacingTabs = new Int32[] { 10, 30, 40, 50, 60, 70, 80, 90 };
		private static void AlignConfigurationSpacing(ref string[] config)
		{
			string[] sa = null;
			for (Int32 i = 0; i <= config.Length - 1; i++) {
				sa = config[i].Split(',');
				config[i] = sa[0].Trim();
				for (Int32 j = 1; j <= sa.Length - 1; j++) {
                    config[i] += "," + Strings.StrDup(Math.Max(0, Module_SaveLoad.SpacingTabs[j - 1] - config[i].Length), " ") + sa[j].Trim();
				}
			}
		}


        public static void Load_ConfigDatabase()
		{
            ThereminoSystem.ConfigDatabase = new object[0]; 
			string[] Config = new string[-1 + 1];
			// ------------------------------------------------------------- 
			string l = null;
			string iniFileName = PlatformAdjustedFileName(Application.StartupPath + "\\" + AssemblyName() + "_ConfigDatabase.txt", "");
			if (Theremino_HAL.My.MyProject.Computer.FileSystem.FileExists(iniFileName)) {
				System.IO.StreamReader f = null;
				f = System.IO.File.OpenText(iniFileName);

				while (!f.EndOfStream) 
                {
					l = f.ReadLine();
					if (!string.IsNullOrEmpty(l)) 
                    {
						// ------------------------------------- store the previous configuration
						if (l.StartsWith("Master,")) 
                        {
                            ThereminoSystem.AddToConfigDatabase(Config);
							Config = new string[-1 + 1];
						}
						if (l.StartsWith("Master,") | l.StartsWith("Slave,") | l.StartsWith("Pin,")) 
                        {
							Array.Resize(ref Config, Config.Length + 1);
							Config[Config.Length - 1] = l;
						}
					}
				}
				f.Close();
				// --------------------------------------------- store the last collected configuration
                ThereminoSystem.AddToConfigDatabase(Config);
			}
		}


	}
}
