using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
namespace Theremino_HAL
{

    static class ThereminoSystem
    {

	    public static bool ConfigurationIsValid = false;

        public static List<Master> Masters;
        public static void TheSystem_InitMasters()
	    {
	        if (Masters != null) 
            {
		        //For Each m As Master In Masters
		        //    '    If m.Hid.HidDevice_Unix.IsHidOpen() Then
		        //    '        m.Hid.HidDevice_Unix.HidClose()
		        //    '    End If
		        //    'm.Hid = Nothing
		        //Next
		        Masters.Clear();
	        }
	        else 
            {
		        Masters = new List<Master>();
	        }

            if (Module_SaveLoad.CanUseWindowsDlls())
            {
			    Theremino_HID.FindThereminoHids();
		    }
		    else 
            {
                Theremino_HID.FindThereminoHids_Unix();
		    }

            // -------------------------------------------- read Master names
            Int32 i = 0;
            while (((i >= 0) && (i < Masters.Count)))
            {
                //  --------------------------------------- read name two times to detect if the Master is already used
                Masters[i].ReadNameFromHardware();
                string name1 = Masters[i].GetName();
                Masters[i].ReadNameFromHardware();
                string name2 = Masters[i].GetName();
                //  --------------------------------------- if a Master is used by another HAL then release it 
                if ((name1 != name2))
                {
                    Masters.Remove(Masters[i]);
                    i--;
                }
                i++;
            }
	    }

	    //Private Function IncreaseFileNameIndex(ByVal filename As String) As String
	    //    If filename = "" Then Return "Noname_01"
	    //    '
        //    Dim path As String = System.IO.Path.GetDirectoryName(filename)
        //    Dim name As String = System.IO.Path.GetFileNameWithoutExtension(filename)
        //    Dim ext As String = System.IO.Path.GetExtension(filename)
	    //    '
	    //    If name = "" Then name = "NoName"
	    //    ' ------------------------------------------------ find index chars
	    //    Dim indexlen As Int32 = 0
	    //    Dim i As Int32 = name.Length - 1
	    //    While Char.IsNumber(name(i)) And i > 0
	    //        i -= 1
	    //        indexlen += 1
	    //    End While
	    //    ' ------------------------------------------------ increase index value
	    //    If indexlen > 0 Then
	    //        Dim number As Int32 = CInt(Val(Mid(name, i + 2)))
	    //        number += 1
	    //        name = Left(name, i + 1) & number.ToString(StrDup(indexlen, "0"))
	    //    Else
	    //        ' -------------------------------------------- or set index to "_01"
	    //        name = name & "_01"
	    //    End If
	    //    ' ------------------------------------------------ compose final file-name
	    //    Return PlatformAdjustedFileName(path & "\" & name & ext)
	    //End Function

        public static void TheSystem_StartTimers()
	    {
            if (!ThereminoSystem.ConfigurationIsValid) return;
 
		    foreach (Master m in Masters) 
            {
			    m.StartTimedOperations();
		    }
	    }

        public static void TheSystem_StopTimers()
	    {
		    foreach (Master m in Masters) 
            {
			    m.StopTimedOperations();
		    }
	    }

        public static void TheSystem_RecognizeSlaves()
	    {
		    foreach (Master m in Masters) 
            {
			    m.Recognize();
		    }

		    AssignDefaultSlotsToPins();
		    AssignConfigurationsToMasters();
		    SetConfigurationParamsToDevices();

		    My.MyProject.Forms.Form1.SelectedLine = -1;
		    TheSystem_ListComponents();
		    My.MyProject.Forms.Form1.ListView_SetAllLineColors();
		    MarkErrorLines();
            My.MyProject.Forms.Form1.Menu_File_Validate.Enabled = !ThereminoSystem.ConfigurationIsValid;
            My.MyProject.Forms.Form1.ToolStripButton_Validate.Enabled = !ThereminoSystem.ConfigurationIsValid;
		    My.MyProject.Forms.Form1.EnableCalibrateButton(TheSystem_CalibrationNeeded());
		    if (IsSnifferPresent()) 
            {
			    My.MyProject.Forms.Form2.btn_TextLog.Enabled = true;
		    }
		    else 
            {
			    My.MyProject.Forms.Form2.btn_TextLog.Enabled = false;
			    My.MyProject.Forms.Form2.btn_TextLog.Checked = false;
		    }

		    TheSystem_Calibrate();
	    }

	    private static void AssignDefaultSlotsToPins()
	    {
		    Int32 i = 1;
		    foreach (Master m in Masters) 
            {
			    Int32 j = i;
			    i += 100;
			    foreach (Slave s in m.Slaves) 
                {
				    foreach (Pin p in s.Pins) 
                    {
					    p.Slot = j;
					    j += 1;
				    }
			    }
		    }
	    }

	    private static bool IsSnifferPresent()
	    {
		    foreach (Master m in Masters) 
            {
			    foreach (Slave s in m.Slaves) 
                {
				    foreach (Pin p in s.Pins) 
                    {
					    if (p.GetPinType() == Pin.PinTypes.SNIFFER) 
                        {
						    return true;
					    }
				    }
			    }
		    }
		    return false;
	    }

        public static bool TheSystem_CalibrationNeeded()
	    {
		    foreach (Master m in Masters) 
            {
			    if (m.CalibrationNeeded()) return true; 
		    }
            return false;
	    }

        public static void TheSystem_Calibrate()
	    {
		    foreach (Master m in Masters) 
            {
			    m.Calibrate();
		    }
	    }

        public static void TheSystem_CalibrateRawZero()
	    {
		    foreach (Master m in Masters) 
            {
			    m.CalibrateRawZero();
		    }
	    }

        public static void TheSystem_SetParkingPositions()
	    {
		    foreach (Master m in Masters) 
            {
			    m.SetOutputPinsToParkingPosition();
		    }
	    }

        public static Int32 TheSystem_GetSlaveCount()
	    {
		    Int32 sc = 0;
		    foreach (Master m in Masters) 
            {
			    sc += m.SlaveCount;
		    }
		    return sc;
	    }

        public static void TheSystem_StopAllSnifferLogs()
	    {
		    foreach (Master m in Masters) 
            {
			    m.StopAllSnifferLogs();
		    }
	    }


        public static Master GetMasterByListLine(Int32 line, string LineType)
	    {
		    Master m = null;
		    Slave s = null;
		    Pin p = null;
		    //
		    switch (LineType) 
            {
			    case "Master":
                    m = ThereminoSystem.FindMasterByListLine(line);
				    break;
			    case "Slave":
                    s = ThereminoSystem.FindSlaveByListLine(line);
				    if (s != null) m = Masters[s.MasterId];
				    break;
			    case "Pin":
                    p = ThereminoSystem.FindPinByListLine(line);
                    if (p != null) m = Masters[p.MasterId];
				    break;
		    }
		    //
		    return m;
	    }

	    private static Master FindMasterByListLine(Int32 line)
	    {
		    Int32 n = 0;
		    foreach (Master m in Masters) 
            {
			    if (n == line) return m; 
			    n += 1;
			    foreach (Slave s in m.Slaves) 
                {
				    n += 1;
				    foreach (Pin p in s.Pins) 
                    {
					    n += 1;
				    }
			    }
		    }
		    return null;
	    }

	    private static Slave FindSlaveByListLine(Int32 line)
	    {
		    Int32 n = 0;
		    foreach (Master m in Masters) 
            {
			    n += 1;
			    foreach (Slave s in m.Slaves) 
                {
				    if (n == line) return s; 
				    n += 1;
				    foreach (Pin p in s.Pins) 
                    {
					    n += 1;
				    }
			    }
		    }
		    return null;
	    }

        public static Pin FindPinByListLine(Int32 line)
	    {
		    Int32 n = 0;
		    foreach (Master m in Masters) 
            {
			    n += 1;
			    foreach (Slave s in m.Slaves) 
                {
				    n += 1;
				    foreach (Pin p in s.Pins) 
                    {
					    if (n == line) return p; 
					    n += 1;
				    }
			    }
		    }
		    return null;
	    }

        public static Pin GetSelectedPin()
	    {
		    return FindPinByListLine(My.MyProject.Forms.Form1.SelectedLine);
	    }



	    // ==================================================================================================
	    //   LIST COMPONENTS
	    // ==================================================================================================

        public static void TheSystem_ListComponents()
	    {
		    ListViewItem item = null;
		    {
			    My.MyProject.Forms.Form1.MyListView1.GridLines = true;
			    My.MyProject.Forms.Form1.MyListView1.FullRowSelect = true;
			    My.MyProject.Forms.Form1.MyListView1.Header_BackColor1 = Color.FromArgb(255, 255, 200);
			    My.MyProject.Forms.Form1.MyListView1.Header_BackColor2 = Color.FromArgb(200, 200, 200);
			    My.MyProject.Forms.Form1.MyListView1.Header_ForeColor = Color.FromArgb(20, 20, 20);
			    My.MyProject.Forms.Form1.MyListView1.Header_ShadowColor = Color.FromArgb(255, 255, 255);
                My.MyProject.Forms.Form1.MyListView1.Header_Font = new Font("Tahoma", 9, FontStyle.Regular);
			    My.MyProject.Forms.Form1.MyListView1.Font = new Font("Courier new", 8);
			    My.MyProject.Forms.Form1.MyListView1.View = View.Details;
			    My.MyProject.Forms.Form1.MyListView1.Clear();
			    // ----------------------------------------------------------------- column headers
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("Type", 56, HorizontalAlignment.Left);
			    //55
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("ID", 28, HorizontalAlignment.Left);
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("Subtype", 105, HorizontalAlignment.Left);
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("Dir.", 36, HorizontalAlignment.Left);
			    //35
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("Slot", 35, HorizontalAlignment.Center);
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("Value", 62, HorizontalAlignment.Right);
			    My.MyProject.Forms.Form1.MyListView1.Columns.Add("Notes", 200, HorizontalAlignment.Left);
			    My.MyProject.Forms.Form1.MyListView1.Refresh();
			    // -----------------------------------------------------------------
                if (Module_SaveLoad.CanUseWindowsDlls())
                {
				    My.MyProject.Forms.Form1.MyListView1.SetExtStyle_DoubleBuffering();
				    My.MyProject.Forms.Form1.MyListView1.BeginUpdate();
			    }
			    foreach (Master m in Masters) 
                {
				    item = My.MyProject.Forms.Form1.MyListView1.Items.Add("Master");
				    item.SubItems.Add((m.MasterId + 1).ToString());
				    item.SubItems.Add(m.GetName());
				    item.SubItems.Add("");
				    item.SubItems.Add("");
				    item.SubItems.Add("");
				    item.SubItems.Add("");
				    foreach (Slave s in m.Slaves) 
                    {
					    item = My.MyProject.Forms.Form1.MyListView1.Items.Add(" Slave");
					    item.SubItems.Add((s.SlaveId + 1).ToString());
					    item.SubItems.Add(s.SlaveTypeString());
					    item.SubItems.Add("");
					    item.SubItems.Add("");
					    item.SubItems.Add("");
					    item.SubItems.Add("");
					    foreach (Pin p in s.Pins) 
                        {
						    item = My.MyProject.Forms.Form1.MyListView1.Items.Add("   Pin");
						    item.SubItems.Add((p.PinId + 1).ToString());
						    item.SubItems.Add(Pin.PinTypeToString(p.GetPinType()));
						    item.SubItems.Add(p.GetDirectionString());
						    if (p.GetPinType() != Pin.PinTypes.UNUSED) 
                            {
							    item.SubItems.Add(p.Slot.ToString());
							    item.SubItems.Add("0.0");
						    }
						    else 
                            {
							    item.SubItems.Add("");
							    item.SubItems.Add("");
						    }
						    item.SubItems.Add("");
					    }
				    }
			    }
			    TheSystem_ListViewResize();
                if (Module_SaveLoad.CanUseWindowsDlls())
                {
				    My.MyProject.Forms.Form1.MyListView1.EndUpdate();
			    }
		    }
	    }

	    private static void MarkErrorLine(Int32 LineNumber)
	    {
		    {
			    My.MyProject.Forms.Form1.MyListView1.Items[LineNumber].SubItems[6].Text = "Found but not configured";
			    My.MyProject.Forms.Form1.MyListView1.Items[LineNumber].BackColor = Color.FromArgb(255, 150, 150);
			    My.MyProject.Forms.Form1.MyListView1.Items[LineNumber].ForeColor = Color.Black;
		    }
	    }

	    private static void AddErrorLine(string[] config, Int32 LineNumber)
	    {
		    {
			    ListViewItem item = null;
			    string[] s = config[LineNumber].Split(',');
			    item = My.MyProject.Forms.Form1.MyListView1.Items.Add(s[0]);
			    item.SubItems.Add("");
			    item.SubItems.Add(s[1]);
			    item.SubItems.Add("");
			    item.SubItems.Add("");
			    item.SubItems.Add("");
			    item.SubItems.Add("Configured but not found");
			    item.BackColor = Color.FromArgb(255, 80, 50);
			    item.ForeColor = Color.White;
		    }
	    }

        public static void TheSystem_ListViewResize()
	    {
		    ResizeLastColumn();
	    }

	    private static void ResizeLastColumn()
	    {
		    if (My.MyProject.Forms.Form1.MyListView1.Columns.Count > 1) 
            {
			    // --------------------------------------------------------- last column
			    Int32 w = 0;
			    for (Int32 i = 0; i <= My.MyProject.Forms.Form1.MyListView1.Columns.Count - 2; i++) {
				    w += My.MyProject.Forms.Form1.MyListView1.Columns[i].Width;
			    }
                w = My.MyProject.Forms.Form1.MyListView1.ClientSize.Width - w - 2;
                if (!ConfigurationIsValid)
                {
                    w = Math.Max(w, 200);
                }
                My.MyProject.Forms.Form1.MyListView1.Columns[My.MyProject.Forms.Form1.MyListView1.Columns.Count - 1].Width = w;
		    }
	    }



	    // ==================================================================================================
	    //   LIST COMPONENTS - UPDATE VALUES
	    // ==================================================================================================

        public static void TheSystem_UpdateListedSubtypes()
	    {
		    string str = null;
		    {
                if (Module_SaveLoad.CanUseWindowsDlls())
                {
				    My.MyProject.Forms.Form1.MyListView1.SetExtStyle_DoubleBuffering();
				    My.MyProject.Forms.Form1.MyListView1.BeginUpdate();
			    }
			    Int32 n = 0;
			    foreach (Master m in Masters) 
                {
				    if (m.GetName() != My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[2].Text) My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[2].Text = m.GetName(); 
				    n += 1;
				    foreach (Slave s in m.Slaves) 
                    {
					    str = s.SlaveTypeString();
					    if (str != My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[2].Text) My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[2].Text = str; 
					    n += 1;
					    foreach (Pin p in s.Pins) 
                        {
						    str = Pin.PinTypeToString(p.GetPinType());
						    if (str != My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[2].Text) My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[2].Text = str; 
						    n += 1;
					    }
				    }
			    }
                if (Module_SaveLoad.CanUseWindowsDlls())
                {
				    My.MyProject.Forms.Form1.MyListView1.EndUpdate();
			    }
		    }
	    }


        public static void TheSystem_UpdateListedValues()
	    {
		    System.Globalization.CultureInfo ci = new System.Globalization.CultureInfo("en-GB");
		    string str = null;
		    {
                if (Module_SaveLoad.CanUseWindowsDlls())
                {
				    // -----------------------------------------------------------------------------
				    //  DoubleBuffering must be set also here ( not only in the list-view creation )
				    // -----------------------------------------------------------------------------
				    My.MyProject.Forms.Form1.MyListView1.SetExtStyle_DoubleBuffering();
				    // -----------------------------------------------------------------------------
				    My.MyProject.Forms.Form1.MyListView1.BeginUpdate();
			    }
			    Int32 n = 0;
			    foreach (Master m in Masters) 
                {
				    if (!m.Hid.MyDeviceDetected) 
                    {
					    My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[6].Text = "DISCONNECTED";
					    My.MyProject.Forms.Form1.MyListView1.Items[n].BackColor = Color.FromArgb(255, 150, 150);
					    My.MyProject.Forms.Form1.MyListView1.Items[n].ForeColor = Color.Black;
				    }
				    n += 1;
				    foreach (Slave s in m.Slaves) 
                    {
					    n += 1;
					    foreach (Pin p in s.Pins) 
                        {
						    if (p.GetPinType() != Pin.PinTypes.UNUSED) {
							    if (float.IsNaN(p.Value)) 
                                {
								    str = "Sleep";
							    }
							    //str = DecodeNan(p.Value)
							    else 
                                {
								    str = p.Value.ToString("0.0", ci);
							    }
							    if (str != My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[5].Text) My.MyProject.Forms.Form1.MyListView1.Items[n].SubItems[5].Text = str; 
						    }
						    n += 1;
					    }
				    }
			    }
                if (Module_SaveLoad.CanUseWindowsDlls())
                {
				    My.MyProject.Forms.Form1.MyListView1.EndUpdate();
			    }
		    }
	    }

	    //Private Function DecodeNan(ByVal n As Single) As String
        //    Dim ns As Single = MemoryMappedFiles.NAN_Sleep
        //    Dim nz As Single = MemoryMappedFiles.NAN_Zero
	    //    Select Case TryCast(n, Object)
	    //        Case CType(ns, Int64)
	    //            Return "Sleep"
	    //        Case CType(nz, Int64)
	    //            Return "Nan_0"
	    //        Case Else
	    //            Return Ctype(n, int64).ToString
	    //    End Select
	    //End Function


	    // ==================================================================================================
	    //   CONFIGURATIONS
	    // ==================================================================================================

	    public static object[] ConfigDatabase;
	    private static void MarkErrorLines()
	    {
		    string[] config = new string[-1 + 1];
		    //

		    foreach (Master m in Masters) 
            {
			    if (m.ConfigValid) continue; 

			    Int32 n = 0;
			    string[] sa = null;
			    //
			    if (ConfigDatabase.Length > 0) 
                {
				    config = ConfigDatabase[m.ConfigId] as string[];
			    }

			    // TODO
			    // if already axists another master with the same name then
			    //m.ConfigValid = False


			    if (n < config.Length) 
                {
				    sa = config[n].Split(',');
				    if (sa.Length < 2 || sa[0].Trim() != "Master" || sa[1].Trim() != m.GetName()) {
					    MarkErrorLine(n);
				    }
			    }
			    else {
				    MarkErrorLine(n);
			    }
			    n += 1;
			    foreach (Slave s in m.Slaves) 
                {
				    if (n < config.Length) 
                    {
					    sa = config[n].Split(',');
					    if (sa.Length < 2 || sa[0].Trim() != "Slave" || sa[1].Trim() != s.SlaveTypeString()) 
                        {
						    MarkErrorLine(n);
					    }
				    }
				    else {
					    MarkErrorLine(n);
				    }
				    n += 1;
				    foreach (Pin p in s.Pins) 
                    {
					    if (n < config.Length) 
                        {
						    sa = config[n].Split(',');
						    if (sa.Length < 2 || sa[0].Trim() != "Pin" || sa[1].Trim() != Pin.PinTypeToString(p.GetPinType())) 
                            {
							    MarkErrorLine(n);
						    }
					    }
					    else 
                        {
						    MarkErrorLine(n);
					    }
					    n += 1;
				    }
			    }
			    //
			    while ((n < config.Length)) 
                {
				    AddErrorLine(config, n);
				    n += 1;
			    }
		    }
	    }

	    private static Int32 CountConfigErrors(Master m, string[] Config)
	    {
		    Int32 Errors = 0;
		    Int32 n = 0;
		    string[] sa = null;
		    //
		    if (n < Config.Length) 
            {
			    sa = Config[n].Split(',');
			    //OrElse sa[1].Trim <> m.Name Then
			    if (sa.Length < 2 || sa[0].Trim() != "Master") 
                {
				    Errors += 1;
			    }
		    }
		    else {
			    Errors += 1;
		    }
		    n += 1;
		    foreach (Slave s in m.Slaves) 
            {
			    if (n < Config.Length)
                {
				    sa = Config[n].Split(',');
				    if (sa.Length < 2 || sa[0].Trim() != "Slave" || sa[1].Trim() != s.SlaveTypeString()) 
                    {
					    Errors += 1;
				    }
			    }
			    else {
				    Errors += 1;
			    }
			    n += 1;
			    foreach (Pin p in s.Pins) 
                {
				    if (n < Config.Length) 
                    {
					    sa = Config[n].Split(',');
					    if (sa.Length < 2 || sa[0].Trim() != "Pin") 
                        {
						    Errors += 1;
					    }
				    }
				    else 
                    {
					    Errors += 1;
				    }
				    n += 1;
			    }
		    }
		    //
		    if (Config.Length > n) 
            {
			    Errors += (Config.Length - n);
		    }
		    //
		    return Errors;
	    }

	    public static string GetConfigName(string[] config)
	    {
		    string[] sa = null;
		    sa = config[0].Split(',');
		    if (sa.Length > 1) 
            {
			    return sa[1].Trim();
		    }
		    else 
            {
			    return "";
		    }
	    }

	    public static Int32 FindConfigByName(string name)
	    {
		    for (Int32 i = 0; i <= ConfigDatabase.Length - 1; i++) 
            {
			    if (name.ToUpper() == GetConfigName(ConfigDatabase[i] as string[]).ToUpper()) 
                {
				    return i;
			    }
		    }
		    return -1;
	    }

        public static void FillConfigNamesCombo(ComboBox cmb)
        {
            for (Int32 i = 0; i <= ConfigDatabase.Length - 1; i++)
            {
                cmb.Items.Add(GetConfigName(ConfigDatabase[i] as string[]));
            }
        }

	    private static void AssignConfigurationsToMasters()
	    {
		    //
            ThereminoSystem.ConfigurationIsValid = true;
		    //
		    for (Int32 i = 0; i <= Masters.Count - 1; i++) 
            {
			    Masters[i].ConfigValid = true;
			    // ----------------------------------------------------------- find configuration by name
			    //m.ConfigId = -1
			    //For j As Int32 = 0 To ConfigDatabase.Length - 1
			    //    If Masters(j).GetName = GetConfigName(TryCast(ConfigDatabase[j], String())) Then
			    //        Masters(j).ConfigId = i
			    //    End If
			    //Next
			    Masters[i].ConfigId = FindConfigByName(Masters[i].GetName());
			    // ----------------------------------------------------------- if not found add a new configuration
			    if (Masters[i].ConfigId < 0) 
                {
				    Masters[i].ConfigId = ConfigDatabase.Length;
				    string[] config = CreateConfig_AsStringArray(Masters[i]);
                    AddToConfigDatabase(config);
                    TheSystem_SaveConfigDatabase();
			    }
			    // ----------------------------------------------------------- test config errors
			    if (CountConfigErrors(Masters[i], ConfigDatabase[Masters[i].ConfigId] as string[]) > 0) 
                {
				    Masters[i].ConfigValid = false;
                    ConfigurationIsValid = false;
			    }
		    }
	    }

	    private static void SetConfigurationParamsToDevices()
	    {
		    //bool PinTypeChanged = false;
		    string[] sa = null;
		    foreach (Master m in Masters) 
            {
			    if (m.ConfigId >= ConfigDatabase.Length) continue; 
			    string[] config = ConfigDatabase[m.ConfigId] as string[];
			    if (config == null) continue; 
			    Int32 n = 0;
			    if (n >= config.Length) continue; 
			    sa = config[n].Split(',');
			    n += 1;
			    if (sa.Length == 4) 
                {
				    m.SetSpeed((int)Conversion.Val(sa[2].Trim()));
				    m.FastComm = sa[3].Trim() == "True";
			    }
			    foreach (Slave s in m.Slaves) 
                {
				    if (n >= config.Length) continue; 
				    sa = config[n].Split(',');
				    n += 1;
				    //If sa.Length >= 2 Then
				    //End If
				    //PinTypeChanged = false;
				    foreach (Pin p in s.Pins) 
                    {
					    if (n >= config.Length) continue; 
					    sa = config[n].Split(',');
					    n += 1;

					    Pin.PinTypes ConfPinType = Pin.StringToPinType(sa[1]);

					    if (ConfPinType != p.GetPinType()) 
                        {
						    s.ConfigurePin(p.PinId, ConfPinType);
						    //PinTypeChanged = true;
					    }

					    if (sa.Length >= 6) {
						    p.Slot = (int)Conversion.Val(sa[2].Trim());
						    if (p.Slot > 999) p.Slot = 999; 
						    p.Value_Max = (float)Conversion.Val(sa[3].Trim());
						    p.Value_Min = (float)Conversion.Val(sa[4].Trim());
						    p.ResponseSpeed = (int)Conversion.Val(sa[5].Trim());
						    if (p.ResponseSpeed > 100) p.ResponseSpeed = 100; 
						    switch (p.GetPinType()) {
							    case Pin.PinTypes.SERVO_8:
							    case Pin.PinTypes.SERVO_16:
								    if (sa.Length >= 8)
                                    {
									    p.MaxTime = (float)Conversion.Val(sa[6].Trim());
									    p.MinTime = (float)Conversion.Val(sa[7].Trim());
								    }

								    break;
							    case Pin.PinTypes.PWM_8:
							    case Pin.PinTypes.PWM_16:
								    if (sa.Length >= 9) {
									    p.MaxTime = (float)Conversion.Val(sa[6].Trim());
									    p.MinTime = (float)Conversion.Val(sa[7].Trim());
									    p.LogResponse = sa[8].Trim() == "True";
								    }

								    break;
							    case Pin.PinTypes.CAP_8:
							    case Pin.PinTypes.CAP_16:
								    if (sa.Length >= 8)
                                    {
									    p.MinVariation = (float)Conversion.Val(sa[6].Trim());
									    p.ProportionalArea = (float)Conversion.Val(sa[7].Trim());
								    }

								    break;
							    case Pin.PinTypes.USOUND_SENSOR:
								    if (sa.Length >= 9)
                                    {
									    p.MaxDist_mm = (float)Conversion.Val(sa[6].Trim());
									    p.MinDist_mm = (float)Conversion.Val(sa[7].Trim());
                                        p.RemoveErrors = sa[8].Trim() == "True";
								    }

								    break;
							    case Pin.PinTypes.CAP_SENSOR:
								    if (sa.Length >= 9)
                                    {
									    p.MaxDist_mm = (float)Conversion.Val(sa[6].Trim());
									    p.MinDist_mm = (float)Conversion.Val(sa[7].Trim());
									    p.Area_cmq = (float)Conversion.Val(sa[8].Trim());
								    }

								    break;
							    case Pin.PinTypes.COUNTER:
							    case Pin.PinTypes.COUNTER_PU:
							    case Pin.PinTypes.FAST_COUNTER:
							    case Pin.PinTypes.FAST_COUNTER_PU:
							    case Pin.PinTypes.PERIOD:
							    case Pin.PinTypes.PERIOD_PU:
								    if (sa.Length >= 9)
                                    {
									    p.MaxFreq = (float)Conversion.Val(sa[6].Trim());
									    p.MinFreq = (float)Conversion.Val(sa[7].Trim());
									    p.ConvertToFreq = sa[8].Trim() == "True";
								    }

								    break;
						    }
					    }
				    }

				    // -------------------------------------- re-align all the pins 
				    //If PinTypeChanged Then
				    m.SetupSlavePins(s.mSlaveId);
				    //End If
			    }
			    m.SetCommunicationIndexesToPins();
		    }
	    }

	    private static string[] CreateConfig_AsStringArray(Master m)
	    {
		    Int32 i = default(Int32);
		    string[] config = new string[1];
		    config[0] = "Master";
		    config[0] += "," + m.GetName();
		    config[0] += "," + m.CommSpeed.ToString();
		    config[0] += "," + m.FastComm.ToString();
		    foreach (Slave s in m.Slaves) 
            {
			    i = config.Length;
			    Array.Resize(ref config, i + 1);
			    config[i] = "Slave";
			    config[i] += "," + s.SlaveTypeString();
			    foreach (Pin p in s.Pins)
                {
				    i = config.Length;
				    Array.Resize(ref config, i + 1);
				    config[i] = "Pin";
				    config[i] += "," + Pin.PinTypeToString(p.GetPinType());
				    config[i] += "," + p.Slot.ToString();
				    config[i] += "," + p.Value_Max.ToString();
				    config[i] += "," + p.Value_Min.ToString();
				    config[i] += "," + p.ResponseSpeed.ToString();
				    switch (p.GetPinType())
                    {
					    case Pin.PinTypes.SERVO_8:
					    case Pin.PinTypes.SERVO_16:
						    config[i] += "," + p.MaxTime.ToString();
						    config[i] += "," + p.MinTime.ToString();
						    break;
					    case Pin.PinTypes.PWM_8:
					    case Pin.PinTypes.PWM_16:
						    config[i] += "," + p.MaxTime.ToString();
						    config[i] += "," + p.MinTime.ToString();
						    config[i] += "," + p.LogResponse.ToString();
						    break;
					    case Pin.PinTypes.CAP_8:
					    case Pin.PinTypes.CAP_16:
						    config[i] += "," + p.MinVariation.ToString();
						    config[i] += "," + p.ProportionalArea.ToString();
						    break;
					    case Pin.PinTypes.USOUND_SENSOR:
						    config[i] += "," + p.MaxDist_mm.ToString();
						    config[i] += "," + p.MinDist_mm.ToString();
                            config[i] += "," + p.RemoveErrors.ToString();
						    break;
					    case Pin.PinTypes.CAP_SENSOR:
						    config[i] += "," + p.MaxDist_mm.ToString();
						    config[i] += "," + p.MinDist_mm.ToString();
						    config[i] += "," + p.Area_cmq.ToString();
						    break;
					    case Pin.PinTypes.COUNTER:
					    case Pin.PinTypes.COUNTER_PU:
					    case Pin.PinTypes.FAST_COUNTER:
					    case Pin.PinTypes.FAST_COUNTER_PU:
					    case Pin.PinTypes.PERIOD:
					    case Pin.PinTypes.PERIOD_PU:
						    config[i] += "," + p.MaxFreq.ToString();
						    config[i] += "," + p.MinFreq.ToString();
						    config[i] += "," + p.ConvertToFreq.ToString();
						    break;
				    }
			    }
		    }
		    return config;
	    }

	    public static void TheSystem_ValidateConfiguration()
	    {
		    foreach (Master m in Masters)
            {
			    m.ConfigValid = true;
		    }
            ThereminoSystem.ConfigurationIsValid = true;
		    TheSystem_ListComponents();
		    My.MyProject.Forms.Form1.ListView_SetAllLineColors();
            My.MyProject.Forms.Form1.Menu_File_Validate.Enabled = !ThereminoSystem.ConfigurationIsValid;
            My.MyProject.Forms.Form1.ToolStripButton_Validate.Enabled = !ThereminoSystem.ConfigurationIsValid;
		    My.MyProject.Forms.Form1.EnableCalibrateButton(TheSystem_CalibrationNeeded());
	    }

	    public static void AddToConfigDatabase(string[] config)
	    {
		    if (config.Length > 0) 
            {
			    Int32 count = ConfigDatabase.Length;
			    Array.Resize(ref ConfigDatabase, count + 1);
			    ConfigDatabase[count] = config;
		    }
	    }

	    public static void TheSystem_UpdateAndSaveConfigDatabase()
	    {
		    TheSystem_UpdateConfigDatabase();
		    TheSystem_SaveConfigDatabase();
	    }

	    public static void TheSystem_UpdateConfigDatabase()
	    {
            if (ThereminoSystem.ConfigurationIsValid)
            {
			    foreach (Master m in Masters)
                {
				    if (m.ConfigValid) 
                    {
					    ConfigDatabase[m.ConfigId] = CreateConfig_AsStringArray(m);
				    }
			    }
		    }
	    }

	    public static void TheSystem_SaveConfigDatabase()
	    {
            if (ThereminoSystem.ConfigurationIsValid)
            {
			    Module_SaveLoad.Save_ConfigDatabase();
		    }
		    else
            {
			    //MsgBox("Invalid configuration - ""Config Database"" not saved.", MsgBoxStyle.Information)
		    }
	    }

    }
}
